/* Definitions and headers for communication with NeXT/Open/GNUstep API.
   Copyright (C) 1989, 1993, 2005 Free Software Foundation, Inc.

This file is part of GNU Emacs.

GNU Emacs is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU Emacs is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU Emacs; see the file COPYING.  If not, write to
the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
Boston, MA 02110-1301, USA.  */

//#include "config.h"
#include "nsgui.h"
#include "frame.h"

#ifdef HAVE_NS

/* NS does not use interrupt-driven input processing (yet), so this is
   unneeded and moreover was causing problems. */
#ifndef COCOA_EXPERIMENTAL_CTRL_G
#undef BLOCK_INPUT
#define BLOCK_INPUT
#undef UNBLOCK_INPUT
#define UNBLOCK_INPUT
#endif

#ifdef __OBJC__

// ============================================================================
//
//   The Emacs application
//
// ============================================================================

// We override sendEvent: as a means to stop/start the event loop
@interface EmacsApp : NSApplication
{
}
- (void)sendEvent:(NSEvent *)theEvent;
- (void)showPreferencesWindow:(id)sender;
@end


// ============================================================================
//
//   The main Emacs view
//
// ============================================================================

@class EmacsToolbar;

@interface EmacsView : NSView <NSTextInput>
   {
   struct frame *emacsframe;
   char *old_title;
   BOOL windowClosing;
   NSString *workingText;
   BOOL processingCompose;
@public
   int rows, cols;
   int scrollbarsNeedingUpdate;
   EmacsToolbar *toolbar;
   }

// AppKit-side interface
- menuDown:sender;
- toolbarClicked: (id)item;
- toggleToolbar: (id)sender;
- (void)keyDown:(NSEvent *)theEvent;
- (void)mouseDown:(NSEvent *)theEvent;
- (void)mouseUp:(NSEvent *)theEvent;
- setMiniwindowImage;

// Emacs-side interface
- initFrameFromEmacs: (struct frame *) f;
- (void) setRows: (int) r andColumns: (int) c;
- (void) setWindowClosing: (BOOL)closing;
- (EmacsToolbar *) toolbar;
- (void) deleteWorkingText;
@end


// ============================================================================
//
//   The main menu implementation (note: menu code needs cleanup)...
//
// ============================================================================

@interface EmacsMenu : NSMenu
   {
     struct frame *frame;
     unsigned long keyEquivModMask;
   }

- initWithTitle:(NSString *)title frame:(struct frame *)f;
- (void)setFrame:(struct frame *)f;
- (void)menuNeedsUpdate:(NSMenu *)menu; // (delegate method)
- (NSString *)parseKeyEquiv:(char *)key;
- (id <NSMenuItem>)addItemWithWidgetValue:(void *)wvptr;
- (void)fillWithWidgetValue:(void *)wvptr;
- (EmacsMenu *)addSubmenuWithTitle:(char *)title forFrame:(struct frame *)f;
- (void) clear;
- (Lisp_Object)runMenuAt:(NSPoint)p forFrame:(struct frame *)f
                 keymaps:(int)keymaps;
@end


// ============================================================================
//
//   Toolbar
//
// ============================================================================

@class EmacsImage;

@interface EmacsToolbar : NSToolbar
   {
     EmacsView *emacsView;
     NSMutableDictionary *identifierToItem;
     NSMutableArray *activeIdentifiers;
     NSArray *prevIdentifiers;
   }
- initForView: (EmacsView *)view withIdentifier: (NSString *)identifier;
- (void) clearActive;
- (BOOL) changed;
- (void) addDisplayItemWithImage: (EmacsImage *)img idx: (int)idx helpText: (char *)help
                         enabled:(BOOL)enabled;
// delegate methods
- (NSToolbarItem *)toolbar: (NSToolbar *)toolbar
     itemForItemIdentifier:(NSString *)itemIdentifier
 willBeInsertedIntoToolbar:(BOOL)flag;
- (NSArray *)toolbarDefaultItemIdentifiers: (NSToolbar *)toolbar;
- (NSArray *)toolbarAllowedItemIdentifiers: (NSToolbar *)toolbar;
@end


// ============================================================================
//
//   Message / question windows
//
// ============================================================================

@interface EmacsDialogPanel : NSPanel
   {
   NSTextField *command;
   NSTextField *title;
   NSMatrix *matrix;
   int rows,cols;
   }
- initFromContents:(Lisp_Object)menu isQuestion:(BOOL)isQ;
- addButton:(char *)str value:(Lisp_Object)val row:(int)row;
- addString:(char *)str row:(int)row;
- addSplit;
- (Lisp_Object)runDialogAt:(NSPoint)p;
@end

@interface EmacsTooltip : NSObject
  {
    NSWindow *win;
    NSTextField *textField;
    NSTimer *timer;
  }
- init;
- (void) setText: (char *)text;
- (void) showAtX: (int)x Y: (int)y for: (int)seconds;
- (void) hide;
- (BOOL) isActive;
- (NSRect) frame;
@end


// ============================================================================
//
// File open/save panels
// This and next override methods to work around OS X behavior of
// restarting application loop when user dismisses panel.
//
// ============================================================================

@interface EmacsSavePanel : NSSavePanel
{
}
@end
@interface EmacsOpenPanel : NSOpenPanel
{
}
@end

@interface EmacsFileDelegate : NSObject
{
}
- (BOOL)panel:(id)sender isValidFilename:(NSString *)filename;
- (BOOL)panel:(id)sender shouldShowFilename:(NSString *)filename;
- (NSString *)panel:(id)sender userEnteredFilename:(NSString *)filename
          confirmed:(BOOL)okFlag;
@end


// ============================================================================
//
//   Images and stippling
//
// ============================================================================

@interface EmacsImage : NSImage
{
  id imageListNext;
  int refCount;
  NSBitmapImageRep *bmRep; // used for accessing pixel data
  unsigned char *pixmapData[5]; // shortcut to access pixel data
  BOOL onTiger;
  NSColor *stippleMask;
}
+ allocInitFromFile:(Lisp_Object)file;
- reference;
- imageListSetNext:(id)arg;
- imageListNext;
- (void)dealloc;
- initFromXBM:(unsigned char *)bits width:(int)w height:(int)h flip:(BOOL)flip;
- initFromSkipXBM:(unsigned char *)bits width:(int)w height:(int)h
             flip:(BOOL)flip length:(int)length;
- initForXPMWithDepth: (int)depth width: (int)width height: (int)height;
- (void)setPixmapData;
- (NSColor *)getPixelAtX: (int)x Y: (int)y;
- (void)setPixelAtX: (int)x Y: (int)y to: (NSColor *)color;
- (void)setAlphaAtX: (int)x Y: (int)y to: (int)a;
- (NSColor *)stippleMask;
@end


// ============================================================================
//
//   Scrollbars
//
// ============================================================================

@interface EmacsScroller : NSScroller
  {
   Lisp_Object win;
   struct frame *frame;
   NSResponder *prevResponder;

   // offset to the bottom of knob of last mouse down
   float last_mouse_offset;
   float min_portion;
   int pixel_height;
   int last_hit_part;

   BOOL condemned;

   // optimize against excessive positioning calls generated by emacs
   int em_position;
   int em_portion;
   int em_whole;
   }

- initFrame:(NSRect )r window:(Lisp_Object)win;
- (void)setFrame:(NSRect)r;
- (void)dealloc;

- setPosition:(int) position portion:(int) portion whole:(int) whole;
- (int) checkSamePosition:(int)position portion:(int)portion whole:(int)whole;
- (void) getMouseMotionPart: (int *)part window: (Lisp_Object *)window
                          x:(Lisp_Object *)x y:( Lisp_Object *)y;
- (void) sendScrollEventAtLoc:(float)loc fromEvent:(NSEvent *)e;
- repeatScroll:(NSTimer *)sender;
- condemn;
- reprieve;
- judge;
@end


// ============================================================================
//
//   Rendering on Panther and above
//
// ============================================================================

#ifdef COCOA
// rendering util
@interface EmacsGlyphStorage : NSObject <NSGlyphStorage>
{
@public
  NSAttributedString *attrStr;
  NSMutableDictionary *dict;
  CGGlyph *cglyphs;
  unsigned long maxChar, maxGlyph;
  long i, len;
}
- initWithCapacity: (unsigned long) c;
- (void) setString: (NSString *)str font: (NSFont *)font;
@end
#endif


// ============================================================================
//
//   Running the preferences window
//
// ============================================================================

@interface EmacsPrefsController : NSObject
{
    struct frame *frame;
    IBOutlet NSWindow *prefsWindow;
    IBOutlet NSPopUpButton *alternateModMenu;
    IBOutlet NSPopUpButton *commandModMenu;
#ifdef COCOA
    IBOutlet NSPopUpButton *controlModMenu;
    IBOutlet NSPopUpButton *functionModMenu;
#endif
    IBOutlet NSMatrix *cursorTypeMatrix;
    IBOutlet NSSlider *cursorBlinkSlider;
    IBOutlet NSSlider *expandSpaceSlider;
#ifdef COCOA
    IBOutlet NSButton *smoothFontsCheck;
    IBOutlet NSButton *useQuickdrawCheck;
    IBOutlet NSButton *useSysHiliteCheck;
    BOOL prevUseHighlightColor;
#endif
    float prevExpandSpace;
    float prevBlinkRate;
}
- (IBAction)cancel:(id)sender;
- (IBAction)ok:(id)sender;
- (IBAction)resetToDefaults:(id)sender;
- (IBAction)runHelp:(id)sender;
- (IBAction)setColors:(id)sender;
- (IBAction)setDefaultFont:(id)sender;

- (void) showForFrame: (struct frame *)f;
- (void) setPanelFromValues;
- (void) setValuesFromPanel;
@end

//extern Lisp_Object ns_menu_path;
extern NSArray *ns_send_types,*ns_return_types;
extern EmacsMenu *mainMenu, *svcsMenu;
#endif  /* __OBJC__ */



// ============================================================================
//
//   Non-OO stuff
//
// ============================================================================

enum ns_cursor_types
{
   no_highlight=0,
   filled_box,
   hollow_box,
   bar,
   line
};


// could use a list to store these, but rest of emacs has a big infrastructure
// for managing a table of bitmap "records"
struct ns_bitmap_record
{
#ifdef __OBJC__
  EmacsImage *img;
#else
  void *img;
#endif
  char *file;
  int refcount;
  int height, width, depth;
};


// init'd in ns_initialize_display_info()
struct ns_display_info
{
  /* Chain of all ns_display_info structures.  */
  struct ns_display_info *next;

  /* This is a cons cell of the form (NAME . FONT-LIST-CACHE).
     The same cons cell also appears in ns_display_name_list.  */
  Lisp_Object name_list_element;

  /* A table of all the fonts we have already loaded.  */
  struct font_info *font_table;

  /* The current capacity of x_font_table.  */
  int font_table_size;

  /* The number of fonts actually stored in x_font_table.
     font_table[n] is used and valid iff 0 <= n < n_fonts.
     0 <= n_fonts <= font_table_size.  */
  int n_fonts;

  /* Minimum width over all characters in all fonts in font_table.  */
  int smallest_char_width;

  /* Minimum font height over all fonts in font_table.  */
  int smallest_font_height;

#ifdef MULTI_KBOARD
  struct kboard *kboard;
#endif

  ///23
  struct ns_bitmap_record *bitmaps;
  int bitmaps_size;
  int bitmaps_last;

  //23
  struct image_cache *image_cache;

  //23: Dimensions and DPI resolution of this screen
  int height, width;
  double resx, resy;

  //23: Mask of things that cause the mouse to be grabbed
  int grabbed;

  //23
  int n_planes;

  //23
  int color_p;

  //23
  Window root_window;

  //23: Xism
  XrmDatabase xrdb;

  //23: The cursor to use for vertical scroll bars.
  Cursor vertical_scroll_bar_cursor;

  //23: most mouse face stuff moved in here (and reasonably so)
  int mouse_face_beg_row, mouse_face_beg_col;
  int mouse_face_end_row, mouse_face_end_col;
  int mouse_face_beg_x, mouse_face_beg_y;
  int mouse_face_end_x, mouse_face_end_y;
  int mouse_face_past_end;
  Lisp_Object mouse_face_window;
  int mouse_face_face_id;
  int mouse_face_deferred_gc;
  Lisp_Object mouse_face_overlay;
  FRAME_PTR mouse_face_mouse_frame;
  int mouse_face_mouse_x, mouse_face_mouse_y;
  int mouse_face_defer;
  int mouse_face_hidden;
  int mouse_face_image_state;

  // these are general, but we redefine due to Xism
  struct frame *ns_highlight_frame;
  struct frame *ns_focus_frame;
#define x_highlight_frame ns_highlight_frame
#define x_focus_frame ns_focus_frame
};

/* This is a chain of structures for all the NS displays currently in use.  */
extern struct ns_display_info *ns_display_list;
// handle Xism
#define x_display_list ns_display_list

extern Lisp_Object ns_display_name_list;
extern struct ns_display_info *ns_display_info_for_name ();

//23: PENDING: these functions (we defined in nsfns) are used in various
// places, but no prototypes are provided
struct ns_display_info *check_x_display_info (Lisp_Object frame);
FRAME_PTR check_x_frame (Lisp_Object frame);


struct ns_output
{
#ifdef __OBJC__
  EmacsView *view;
  id miniimage;
  NSColor *current_cursor_color;
  NSColor *desired_cursor_color;
  NSColor *foreground_color;
  NSColor *background_color;
  EmacsToolbar *toolbar;
#else
  void *view;
  void *miniimage;
  void *current_cursor_color;
  void *desired_cursor_color;
  void *foreground_color;
  void *background_color;
  void *toolbar;
#endif

  //23: NSCursors init'ed in initFrameFromEmacs
  Cursor text_cursor;
  Cursor nontext_cursor;
  Cursor modeline_cursor;
  Cursor hand_cursor;
  Cursor hourglass_cursor;
  Cursor horizontal_drag_cursor;

  //23: NS-specific
  Cursor current_pointer;

  //23: lord knows why Emacs needs to know about our Window ids..
  Window window_desc, parent_desc;
  char explicit_parent;

  XFontStruct *font;

  /* If a fontset is specified for this frame instead of font, this
     value contains an ID of the fontset, else -1.  */
  int fontset; // only used with font_backend

#ifdef USE_FONT_BACKEND
  struct font *fontp;
#endif	/* USE_FONT_BACKEND */

  Lisp_Object icon_top;
  Lisp_Object icon_left;
  enum ns_cursor_types current_cursor, desired_cursor;

  /* The size of the extra width currently allotted for vertical
     scroll bars, in pixels.  */
  int vertical_scroll_bar_extra;

  /* The height of the titlebar decoration (included in NSWindow's frame). */
  int titlebar_height;

  /* The height of the toolbar if displayed, else 0. */
  int toolbar_height;

  /* This is the Emacs structure for the NS display this frame is on.  */
  struct ns_display_info *display_info;
};

//23: this dummy decl now needed to support TTYs
struct x_output
{
  unsigned long background_pixel;
  unsigned long foreground_pixel;
};


/* This gives the ns_display_info structure for the display F is on.  */
#define FRAME_NS_DISPLAY_INFO(f) ((f)->output_data.ns->display_info)
// the primacy of X must be constantly worked with...
#define FRAME_X_DISPLAY_INFO(f) ((f)->output_data.ns->display_info)
#define FRAME_X_OUTPUT(f) ((f)->output_data.ns)
#define FRAME_NS_WINDOW(f) ((f)->output_data.ns->window_desc)
#define FRAME_X_WINDOW(f) ((f)->output_data.ns->window_desc)

/* This is the `Display *' which frame F is on.  */
#define FRAME_NS_DISPLAY(f) (0)
#define FRAME_X_DISPLAY(f) (0)

#define FRAME_FOREGROUND_COLOR(f) ((f)->output_data.ns->foreground_color)
#define FRAME_BACKGROUND_COLOR(f) ((f)->output_data.ns->background_color)
#define FRAME_FOREGROUND_PIXEL(f) FRAME_FOREGROUND_COLOR(f)
#define FRAME_BACKGROUND_PIXEL(f) FRAME_BACKGROUND_COLOR(f)

#define FRAME_X_IMAGE_CACHE(F) FRAME_NS_DISPLAY_INFO ((F))->image_cache

#define NS_FACE_FOREGROUND(f) ((NSColor *)f->foreground)
#define NS_FACE_BACKGROUND(f) ((NSColor *)f->background)
#define FRAME_NS_TITLEBAR_HEIGHT(f) ((f)->output_data.ns->titlebar_height)
#define FRAME_NS_TOOLBAR_HEIGHT(f) ((f)->output_data.ns->toolbar_height)

#ifndef COMPILING_MULTI_TERM
#define FONT_WIDTH(f)	((f)->width)
#define FONT_HEIGHT(f)	((f)->height)
#define FONT_BASE(f)    ((f)->max_bounds.ascent)
#define FONT_DESCENT(f) ((f)->max_bounds.descent)

#define FRAME_DEFAULT_FACE(f) FACE_FROM_ID(f, DEFAULT_FACE_ID)
#endif /* ! COMPILING_MULTI_TERM */

#define FRAME_NS_VIEW(f) ((f)->output_data.ns->view)
#define FRAME_CURSOR(f)  ((f)->output_data.ns->current_cursor)
#define FRAME_CURSOR_COLOR(f) ((f)->output_data.ns->current_cursor_color)
#define FRAME_NEW_CURSOR_COLOR(f) ((f)->output_data.ns->desired_cursor_color)
#define FRAME_NEW_CURSOR(f)  ((f)->output_data.ns->desired_cursor)
#define FRAME_POINTER_TYPE(f) ((f)->output_data.ns->current_pointer)

#ifndef COMPILING_MULTI_TERM
#define FRAME_FONT(f) ((f)->output_data.ns->font)
#endif /* ! COMPILING_MULTI_TERM */

#ifdef USE_FONT_BACKEND
#define FRAME_FONT_OBJECT(f) ((f)->output_data.ns->fontp)
#endif	/* USE_FONT_BACKEND */

#ifdef __OBJC__
#define XNS_SCROLL_BAR(vec) ((id) LISP_TO_VOID (vec))
#else
#define XNS_SCROLL_BAR(vec) LISP_TO_VOID (vec)
#endif

/* Compute pixel size for vertical scroll bars */
#define NS_SCROLL_BAR_WIDTH(f)                              \
(FRAME_HAS_VERTICAL_SCROLL_BARS(f)                          \
 ? rint(FRAME_CONFIG_SCROLL_BAR_WIDTH (f) > 0               \
        ? FRAME_CONFIG_SCROLL_BAR_WIDTH (f)                 \
        : (FRAME_SCROLL_BAR_COLS (f) * FRAME_COLUMN_WIDTH(f)))   \
 : 0)

/* Difference btwn char-column-calculated and actual SB widths.
   This is only a concern for rendering when SB on left. */
#define NS_SCROLL_BAR_ADJUST(w, f)		\
(WINDOW_HAS_VERTICAL_SCROLL_BAR_ON_LEFT(w) ?	\
    (FRAME_SCROLL_BAR_COLS (f) * FRAME_COLUMN_WIDTH(f)	\
        - NS_SCROLL_BAR_WIDTH(f)) : 0)

//PENDING: fix for GNUstep inconsistent accounting for titlebar
#ifdef GNUSTEP
#define NS_TOP_POS(f) ((f)->top_pos + 18)
#else
#define NS_TOP_POS(f) ((f)->top_pos)
#endif

#define FRAME_NS_FONT_TABLE(f) (FRAME_NS_DISPLAY_INFO (f)->font_table)

#ifdef USE_FONT_BACKEND
#define FRAME_FONTSET(f) ((f)->output_data.ns->fontset)
#else
#define FRAME_FONTSET(f) -1
#endif

//23: some of these may need better definitions..
#define FRAME_SMALLEST_CHAR_WIDTH(f)  (((f)->output_data.ns->font)->width)
#define FRAME_SMALLEST_FONT_HEIGHT(f) (((f)->output_data.ns->font)->height)
#define FONT_TYPE_FOR_UNIBYTE(font, ch)   0
#define FONT_TYPE_FOR_MULTIBYTE(font, ch) 0
#define FRAME_BASELINE_OFFSET(f) 0
#define BLACK_PIX_DEFAULT(f) 0x000000
#define WHITE_PIX_DEFAULT(f) 0xFFFFFF

/* First position where characters can be shown (instead of scrollbar, if
   it is on left. */
#define FIRST_CHAR_POSITION(f) (!(FRAME_HAS_VERTICAL_SCROLL_BARS_ON_LEFT(f)) ?\
         0 : FRAME_SCROLL_BAR_COLS(f))

extern struct ns_display_info *ns_term_init ();
extern void ns_term_shutdown(int sig);

// constants for text rendering
#define NS_DUMPGLYPH_NORMAL             0
#define NS_DUMPGLYPH_CURSOR             1
#define NS_DUMPGLYPH_FOREGROUND         2
#define NS_DUMPGLYPH_MOUSEFACE          3


// In nsfont, called from fontset.c
extern void nsfont_make_fontset_for_font(Lisp_Object name,
                                         Lisp_Object font_object);

// In nsfont, for debugging
void dump_glyphstring(struct glyph_string *s);

// Implemented in nsterm, published in or needed from nsfns.
extern Lisp_Object Qfontsize;
extern Lisp_Object ns_new_font();
extern struct font_info *ns_load_font (struct frame *f,char *name,int sz);
extern Lisp_Object ns_list_fonts ();
extern struct font_info *ns_get_font_info ();
extern struct font_info *ns_query_font (struct frame *f, char *fontname);
extern Lisp_Object ns_list_fonts (FRAME_PTR f, Lisp_Object pattern,
                                  int size, int maxnames);
extern void ns_find_ccl_program (struct font_info *fontp);

#ifdef __OBJC__
extern const char *ns_font_to_xlfd(NSFont *font);
#endif
extern const char *ns_fontname_to_xlfd(const char *name);
extern const char *ns_xlfd_to_fontname(const char *xlfd);

extern void check_ns(void);
extern Lisp_Object ns_map_event_to_object ();
extern Lisp_Object ns_string_from_pasteboard();
extern void ns_string_to_pasteboard();
extern void nxatoms_of_nsselect();
extern int ns_lisp_to_cursor_type();
extern Lisp_Object ns_cursor_type_to_lisp(int arg);
extern Lisp_Object Qnone;

// XColor defined in dispextern.h (we use color_def->pixel = NSColor id), but
// this causes an #include snafu, so we can't declare it.
//  extern int ns_defined_color(struct frame *f, char *name, XColor *color_def,
//                              int alloc);

#ifdef __OBJC__
extern int ns_lisp_to_color (Lisp_Object color, NSColor **col);
#endif

// C access to ObjC functionality
extern void  ns_release_object (void *obj);
extern void  ns_retain_object (void *obj);
extern void *ns_alloc_autorelease_pool ();
extern void ns_release_autorelease_pool ();

// in nsmenu
extern void update_frame_tool_bar (FRAME_PTR f);
extern void free_frame_tool_bar (FRAME_PTR f);
extern void find_and_call_menu_selection (FRAME_PTR f, void *client_data);
extern Lisp_Object find_and_return_menu_selection (FRAME_PTR f,
                                                   int keymaps,
                                                   void *client_data);

// two more prototypes that should be moved to a more general include file
extern void set_frame_menubar (struct frame *f, int first_time, int deep_p);
extern void x_set_window_size(struct frame *f, int change_grav,
                              int cols, int rows);

// From nsimage.m, needed in image.c
struct image;
extern void *ns_image_from_XBM (unsigned char *bits, int width, int height);
extern void *ns_image_for_XPM (int width, int height, int depth);
extern void *ns_image_from_file (Lisp_Object file);
extern int ns_load_image (struct frame *f, struct image *img,
                          Lisp_Object spec_file, Lisp_Object spec_data);
extern int ns_image_width (void *img);
extern int ns_image_height (void *img);
extern unsigned long ns_get_pixel(void *img, int x, int y);
extern void ns_put_pixel(void *img, int x, int y, void *color);
extern void ns_set_alpha(void *img, int x, int y, int a);
extern int ns_get_red(void *color);
extern int ns_get_green(void *color);
extern int ns_get_blue(void *color);

/* This in nsterm.m */
extern unsigned long ns_get_rgb_color(float r, float g, float b, float a);
extern NSPoint last_mouse_motion_position;

#ifdef GNUSTEP
extern char gnustep_base_version[];  /* version tracking */
#endif

#define MINWIDTH 10
#define MINHEIGHT 10

/* Screen max coordinate
 Using larger coordinates causes movewindow/placewindow to abort */
#define SCREENMAX 16000

#define NS_SCROLL_BAR_WIDTH_DEFAULT     [EmacsScroller scrollerWidth]
/* This is to match emacs on other platforms, ugly though it is. */
#define NS_SELECTION_COLOR_DEFAULT	@"LightGoldenrod2";

/* Little utility macros */
#define IN_BOUND(min,x,max) (((x)<(min)) ? (min) : (((x)>(max)) ? (max) : (x)))
#define SCREENMAXBOUND(x) (IN_BOUND(-SCREENMAX,x,SCREENMAX))

//23: needed somewhere...
#define VERTICAL_SCROLL_BAR_WIDTH_TRIM (0)


#endif
